from bitcoinrpc._spec import Error, RequestId


class RPCError(Exception):
    """
    Enrich the `Error` - https://www.jsonrpc.org/specification#error_object
    with the `id` of the request that caused the error.
    """

    def __init__(self, id: RequestId, error: Error) -> None:
        super().__init__(error["message"])
        self.id = id
        self.error = error



"""
Module containing definitions of objects (described as Python's Typed dictionaries)
taken from JSON-RPC 2.0 specification.

https://www.jsonrpc.org/specification
"""

from typing import List, Optional, Union

from typing_extensions import NotRequired, TypedDict

from bitcoinrpc._types import JSONType

RequestId = Union[int, str, None]
"""
https://www.jsonrpc.org/specification#request_object

According to the specification, must be String, Integer or Null.
"""


class Request(TypedDict):
    """
    https://www.jsonrpc.org/specification#request_object

    `jsonrpc` member is always "2.0"
    """

    jsonrpc: str
    id: RequestId
    method: str
    params: List[JSONType]


class Error(TypedDict):
    """
    https://www.jsonrpc.org/specification#error_object

    `data` member may not be present at all.
    """

    code: int
    message: str
    data: NotRequired[JSONType]


class _ResponseCommon(TypedDict):
    jsonrpc: str
    id: RequestId


class ResponseSuccess(_ResponseCommon):
    """
    https://www.jsonrpc.org/specification#response_object

    - `jsonrpc` member is always "2.0"
    - `result` member is always present and is a valid JSON
    - `error` should not be present by the specification, but at least Bitcoin Node
        as of version v26.0 always includes `error` member with value `NULL` on the
        success path.
    """

    result: JSONType
    error: NotRequired[Optional[Error]]


class ResponseError(_ResponseCommon):
    """
    https://www.jsonrpc.org/specification#response_object

    - `jsonrpc` member is always "2.0"
    - `result` should not be present by the specification, but at least Bitcoin Node
        as of version v26.0 always includes `result` member with value `NULL` on the
        error path.
    - `error` member is always present and is a JSON of `Error` shape defined above.
    """

    result: NotRequired[Optional[JSONType]]
    error: Error


Response = Union[ResponseSuccess, ResponseError]



import typing as t

import httpx
import orjson
import pytest

from bitcoinrpc import BitcoinRPC


@pytest.mark.asyncio
async def test_two_clients_dont_share_counter(rpc_config: t.Dict[str, t.Any]) -> None:
    logs: t.List[bytes] = []

    def handler(_: httpx.Request) -> httpx.Response:
        """
        Mock the JSON-RPC response so this test does not rely on a running Bitcoin Node
        """
        return httpx.Response(200, json={"jsonrpc": "2.0", "id": 1, "result": 0})

    async def log_request(request: httpx.Request) -> None:
        """
        Hook that stores the serialized request before it is sent over the wire.
        """
        nonlocal logs
        logs.append(orjson.loads(request.content))

    url, auth = rpc_config["url"], rpc_config["auth"]
    client1 = httpx.AsyncClient(
        auth=auth,
        event_hooks={"request": [log_request]},
        transport=httpx.MockTransport(handler=handler),
    )
    client2 = httpx.AsyncClient(
        auth=auth,
        event_hooks={"request": [log_request]},
        transport=httpx.MockTransport(handler=handler),
    )

    async with BitcoinRPC(url=url, client=client1) as rpc1:
        _ = await rpc1.getblockcount()

    async with BitcoinRPC(url=url, client=client2) as rpc2:
        _ = await rpc2.getblockcount()

    req1, req2 = logs
    assert req1 == req2
